int infinity  = 10000;
int undefined = -1;

PImage country;
PImage tower_zero;
PImage tower_one;
PImage tower_two;

PFont font45;
int towers=0;
int maxtowers=10;
Tower[] twr = new Tower[maxtowers]; 
int[][] distBetween;

class Tower { 
  int xloc; // y value of location
  int yloc; // y value of location
  int lvl; // level,  being height
  int typ; // type,  either 0,1, 2
  Tower (int tp) {  
    xloc = 0; 
    yloc = 0;
    lvl = 0;
    typ = tp; 
  } 
  void moveTo(int x,int y) {
    xloc = x;
    yloc = y;
  }
  void setLevel(int lv) {
    lvl = lv;
  }
  void show() { 
    tint(255,128);
    image(getTowerImage(typ),xloc-25,yloc-lvl,50,lvl);
  } 
  void showEmphasised(){
    tint(255,255);
    image(getTowerImage(typ),xloc-25,yloc-lvl,50,lvl);  
  }
}

boolean shownodes=false;
boolean showedges=false;
boolean showpath =false;
int curtyp=0;  // current tower type
int curlvl=50; // proposed new height
    
PImage getTowerImage(int typ) {
  PImage b;
  switch(typ) {
    case 0 : b = tower_zero; break;
    case 1 : b = tower_one;  break;
    case 2 : b = tower_two;  break;
    default: b = tower_zero;
  }
  return b;
}

int[][] mkGraph() {
   int[][] distance = new int[towers][towers];
   for (int i=0; i<towers; i++) {
       for (int j=0; j<i; j++)  {
           // physical distance by pythagoras
           int a=twr[i].xloc-twr[j].xloc;
           int b=twr[i].yloc-twr[j].yloc;
           int c=floor(sqrt(a*a + b*b));
           if (random(0,10) > 4) {
              distance[i][j]=c;
              distance[j][i]=c;
           }  else {
              distance[i][j] = infinity;
              distance[j][i] = infinity;
           }
       distance[i][i]=0;       // point to itself is zero distance
       }
     distance[0][1]=infinity;  // dont make the problem too simple
     distance[1][0]=infinity; 
   }
   return distance;
}

//void dmpGraph() {
//     for (int i=0; i<towers; i++) {
//         for (int j=0; j<towers; j++) {
//             print(i);print(",");print(j);print(":");print(distBetween[i][j]);print(";  ");
//         }
//         println();
//     }
//}

void showNodes() {
   for (int i=0; i<towers; i++) {
       twr[i].showEmphasised();
       fill(0);
       stroke(color(255,0,0));
       if (i<2) strokeWeight(3); else strokeWeight(1);
       ellipse(twr[i].xloc,twr[i].yloc,10,10);   
   }
}   
   
void showEdges() {
   for (int i=0; i<towers; i++)
       for (int j=0; j<towers; j++)
           if (distBetween[i][j]<infinity) {
              stroke(0, 0, 0);
              strokeWeight(1);
              line(twr[i].xloc, twr[i].yloc, twr[j].xloc, twr[j].yloc);
           }
}

void showOrigin() {
   textFont(font45,18);
   fill(255, 0, 0);
   text("from", twr[0].xloc-22, twr[0].yloc-twr[0].lvl);
}

void showDestination() {
   fill(255, 0, 0);
   textFont(font45,18);
   text("to", twr[1].xloc-10, twr[1].yloc-twr[1].lvl);
}

void showPath(int[] previous) {
     // see http://en.wikipedia.org/wiki/Dijkstra_algorithm
     int u = 1; //target
     int undefined = -1;
     while (previous[u]>undefined) {
           stroke(255, 0, 0);
           strokeWeight(3);
           line(twr[previous[u]].xloc, twr[previous[u]].yloc, twr[u].xloc, twr[u].yloc );
           u = previous[u];
     } 
}

// variables and auxiliaries, used in 
// Dijkstra's shortest path algorithm
int[] dIst;
int[] prvs;

boolean isNotEmpty(boolean[] q){
   int i=0;
   boolean isEmpty=true;
   while (i<towers && isEmpty) {
         if (q[i]) isEmpty=false;
         i++;
   }
   return !isEmpty;
}

int nodeWithSmallestDistIn(boolean[] q) {
    int m = infinity; // min level found so far
    int i = -1;       // index where it's found
    for (int j=0; j<towers; j++)
        if (q[j] && dIst[j] <= m) {
           m = dIst[j];
           i = j;
        }
    return i;
}

int[] dijkstra() { 
// see http://en.wikipedia.org/wiki/Dijkstra_algorithm
// takes distBetween as graph and index zero as source
   int undefined = -1;
   dIst = new int[towers];
   int[] previous = new int[towers];
   for (int v=0; v<towers; v++) {        // Initializations
       dIst[v] = infinity;               // Unknown distance function from source to v
       previous[v] = undefined;          // Previous node in optimal path from source
   }
   int source=0;
   dIst[source] = 0;                     // Distance from source to source
   boolean[] Q = new boolean[towers];    // All nodes in the graph are unoptimized -
   for (int i=0;i<towers;i++) {
       Q[i]=true;                        // thus are in Q
   } 
   while (isNotEmpty(Q)) {               // The main loop
         int u = nodeWithSmallestDistIn(Q); 
         Q[u]=false; 
         for (int v=0; v<towers; v++) {  // where v has not yet been removed from Q.
             if (v!=u && distBetween[u][v] < infinity && Q[v]) {
                int alt = dIst[u] + distBetween[u][v];
                if (alt < dIst[v]) {     // Relax (u,v)
                   dIst[v] = alt;
                   previous[v] = u;
                }
             }
         }
    }
    return previous;
}

void setup() {
  size(500,500);
  frameRate(10);
  font45 = loadFont("Univers45.vlw"); 
  // The font must be located in the sketch's 
  // "data" directory to load successfully
  country    = loadImage("country.jpg");
  tower_zero = loadImage("tower_zero.jpg");
  tower_one  = loadImage("tower_one.jpg");
  tower_two  = loadImage("tower_two.jpg");
}

void draw() {
   // display background
   tint(255,128);
   image(country,0,0);
 
   // display cursor and preview tower
   tint(255,128);
   image(getTowerImage(curtyp),mouseX-25,mouseY-curlvl,50,curlvl);

   // display towers for indexes 0..towers-1
   tint(255,128);
   for (int i=0; i<towers; i++) {
       twr[i].showEmphasised();
   }
   // show graph structure 
   if (shownodes) showNodes();
   if (showedges) showEdges(); 
   if (showpath)  showPath(prvs);    
   if (towers>0)  showOrigin();
   if (towers>1)  showDestination();
}

void mousePressed() {
   if (towers<maxtowers && shownodes==false){
      Tower t = new Tower(curtyp);
      t.setLevel(curlvl);
      t.moveTo(mouseX,mouseY);
      twr[towers++] = t;
   }
}

void keyPressed() {
   if(keyCode == LEFT)  { curtyp=(curtyp+1)%3; } 
   if(keyCode == RIGHT) { curtyp=((curtyp+3)-1)%3; } 
   if(keyCode == UP)    { curlvl+=3; } 
   if(keyCode == DOWN)  { if (curlvl>10) curlvl-=3; } 
   if(key == ' ')       { } 
   if (keyCode == ENTER) { 
      if (shownodes==false & towers>1) {
         shownodes=true;
      } else if (showedges==false & towers>1) {
             showedges=true; 
             distBetween=mkGraph();    
      } else if (showpath==false & towers>1) {
             showpath=true;
             prvs=dijkstra();
      }
   }
}


